import json
import warnings
from collections.abc import Iterator
from functools import cached_property
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click
from eip712.messages import EIP712Message
from eth_account import Account as EthAccount
from eth_account._utils.signing import sign_transaction_dict
from eth_account.hdaccount import ETHEREUM_DEFAULT_PATH
from eth_account.messages import encode_defunct
from eth_keys.datatypes import PrivateKey
from eth_pydantic_types import HexBytes
from eth_pydantic_types import HexStr as PydanticHexStr
from eth_typing import HexStr
from eth_utils import remove_0x_prefix, to_bytes, to_canonical_address, to_hex

from ape.api.accounts import AccountAPI, AccountContainerAPI
from ape.api.address import BaseAddress
from ape.exceptions import AccountsError, SignatureError
from ape.logging import logger
from ape.types.address import AddressType
from ape.types.signatures import MessageSignature, SignableMessage, TransactionSignature
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.misc import ZERO_ADDRESS, derive_public_key, log_instead_of_fail
from ape.utils.validators import _validate_account_alias, _validate_account_passphrase
from ape_ethereum import Authorization
from ape_ethereum.transactions import TransactionType

if TYPE_CHECKING:
    from eth_account.signers.local import LocalAccount

    from ape.api.transactions import TransactionAPI


class InvalidPasswordError(AccountsError):
    """
    Raised when password to unlock an account is incorrect.
    """

    def __init__(self):
        super().__init__("Invalid password")


class AccountContainer(AccountContainerAPI):
    loaded_accounts: dict[str, "KeyfileAccount"] = {}

    @property
    def _keyfiles(self) -> Iterator[Path]:
        return self.data_folder.glob("*.json")

    @property
    def aliases(self) -> Iterator[str]:
        for p in self._keyfiles:
            yield p.stem

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for keyfile in self._keyfiles:
            if keyfile.stem not in self.loaded_accounts:
                keyfile_account = KeyfileAccount(keyfile_path=keyfile)
                self.loaded_accounts[keyfile.stem] = keyfile_account

            yield self.loaded_accounts[keyfile.stem]

    def __len__(self) -> int:
        return len([*self._keyfiles])


class ApeSigner(AccountAPI):
    private_key: PydanticHexStr

    @cached_property
    def address(self) -> AddressType:
        return EthAccount.from_key(self.private_key).address

    @property
    def public_key(self) -> "HexBytes":
        return derive_public_key(HexBytes(self.private_key))

    def sign_authorization(
        self,
        address: AddressType,
        chain_id: int | None = None,
        nonce: int | None = None,
    ) -> MessageSignature | None:
        if chain_id is None:
            chain_id = self.chain_manager.chain_id

        signed_authorization = EthAccount.sign_authorization(
            dict(
                chainId=chain_id,
                address=to_canonical_address(address),
                nonce=nonce or self.nonce,
            ),
            self.private_key,
        )
        return MessageSignature(
            v=signed_authorization.y_parity,
            r=to_bytes(signed_authorization.r),
            s=to_bytes(signed_authorization.s),
        )

    def sign_message(self, msg: Any, **signer_options) -> MessageSignature | None:
        # Convert str and int to SignableMessage if needed
        if isinstance(msg, str):
            msg = encode_defunct(text=msg)
        elif isinstance(msg, int):
            msg = to_hex(msg)
            msg = encode_defunct(hexstr=msg)
        elif isinstance(msg, EIP712Message):
            # Convert EIP712Message to SignableMessage for handling below
            msg = msg.signable_message

        # Process SignableMessage
        if isinstance(msg, SignableMessage):
            signed_msg = EthAccount.sign_message(msg, self.private_key)
            return MessageSignature(
                v=signed_msg.v,
                r=to_bytes(signed_msg.r),
                s=to_bytes(signed_msg.s),
            )

        return None

    def sign_transaction(self, txn: "TransactionAPI", **signer_options) -> "TransactionAPI | None":
        # Signs any transaction that's given to it.
        # NOTE: Using JSON mode, as only primitive types can be signed.
        tx_data = txn.model_dump(mode="json", by_alias=True, exclude={"sender"})
        private_key = PrivateKey(HexBytes(self.private_key))

        # NOTE: var name `sig_r` instead of `r` to avoid clashing with pdb commands.
        try:
            (
                sig_v,
                sig_r,
                sig_s,
                _,
            ) = sign_transaction_dict(private_key, tx_data)
        except TypeError as err:
            # Occurs when missing properties on the txn that are needed to sign.
            raise SignatureError(str(err), transaction=txn) from err

        # NOTE: Using `to_bytes(hexstr=to_hex(sig_r))` instead of `to_bytes(sig_r)` as
        #   a performance optimization.
        txn.signature = TransactionSignature(
            v=sig_v,
            r=to_bytes(hexstr=to_hex(sig_r)),
            s=to_bytes(hexstr=to_hex(sig_s)),
        )

        return txn

    def sign_raw_msghash(self, msghash: HexBytes) -> MessageSignature:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            signed_msg = EthAccount.unsafe_sign_hash(msghash, self.private_key)

        return MessageSignature(
            v=signed_msg.v,
            r=to_bytes(signed_msg.r),
            s=to_bytes(signed_msg.s),
        )

    def set_delegate(self, contract: BaseAddress | AddressType | str, **txn_kwargs):
        contract_address = self.conversion_manager.convert(contract, AddressType)
        sig = self.sign_authorization(contract_address, nonce=self.nonce + 1)
        auth = Authorization.from_signature(
            address=contract_address,
            chain_id=self.chain_manager.chain_id,
            # NOTE: `tx` uses `self.nonce`
            nonce=self.nonce + 1,
            signature=sig,
        )
        tx = self.provider.network.ecosystem.create_transaction(
            type=TransactionType.SET_CODE,
            authorizations=[auth],
            sender=self,
            # NOTE: Cannot target `ZERO_ADDRESS`
            receiver=txn_kwargs.pop("receiver", None) or self,
            **txn_kwargs,
        )
        return self.call(tx)

    def remove_delegate(self, **txn_kwargs):
        sig = self.sign_authorization(ZERO_ADDRESS, nonce=self.nonce + 1)
        auth = Authorization.from_signature(
            chain_id=self.chain_manager.chain_id,
            address=ZERO_ADDRESS,
            # NOTE: `tx` uses `self.nonce`
            nonce=self.nonce + 1,
            signature=sig,
        )
        tx = self.provider.network.ecosystem.create_transaction(
            type=TransactionType.SET_CODE,
            authorizations=[auth],
            sender=self,
            # NOTE: Cannot target `ZERO_ADDRESS`
            receiver=txn_kwargs.pop("receiver", None) or self,
            **txn_kwargs,
        )
        return self.call(tx)


# NOTE: `AccountAPI` is an BaseInterfaceModel
class KeyfileAccount(AccountAPI):
    keyfile_path: Path
    locked: bool = True
    __autosign: bool = False
    __cached_signer: ApeSigner | None = None

    @log_instead_of_fail(default="<KeyfileAccount>")
    def __repr__(self) -> str:
        # NOTE: Prevent errors from preventing repr from working.
        address_str = f" address={self.address}" if self.address else ""
        alias_str = f" alias={self.alias}" if self.alias else ""
        return f"<{self.__class__.__name__}{address_str}{alias_str}>"

    @property
    def alias(self) -> str:
        return self.keyfile_path.stem

    @property
    def keyfile(self) -> dict:
        return json.loads(self.keyfile_path.read_text())

    @property
    def address(self) -> AddressType:
        return self.network_manager.ethereum.decode_address(self.keyfile["address"])

    @property
    def __signer(self) -> ApeSigner:
        if self.__cached_signer is not None:
            if not self.locked:
                logger.warning("Using cached key for %s", self.alias)
                return self.__cached_signer

            self.__cached_signer = None

        passphrase = self._prompt_for_passphrase(default="")
        key = self.__decrypt_keyfile(passphrase)
        signer = ApeSigner(private_key=key)

        if click.confirm(f"Leave '{self.alias}' unlocked?"):
            self.locked = False
            self.__cached_signer = signer

        return signer

    @property
    def public_key(self) -> HexBytes | None:
        keyfile_data = self.keyfile
        if "public_key" in keyfile_data:
            return HexBytes(bytes.fromhex(keyfile_data["public_key"]))

        # Derive the public key from the private key
        public_key = self.__signer.public_key
        keyfile_data["public_key"] = remove_0x_prefix(HexStr(public_key.hex()))

        # Store the public key so we don't have to derive it again.
        self.keyfile_path.write_text(json.dumps(keyfile_data), encoding="utf8")

        return public_key

    def unlock(self, passphrase: str | None = None):
        if not passphrase:
            # Check if environment variable is available
            env_variable = f"APE_ACCOUNTS_{self.alias}_PASSPHRASE"
            passphrase = environ.get(env_variable, None)

            if passphrase:
                # Passphrase found in environment variable
                logger.info(
                    f"Using passphrase for account '{self.alias}' from environment variable"
                )
            else:
                # Passphrase not found, prompt for it
                passphrase = self._prompt_for_passphrase(
                    f"Enter passphrase to permanently unlock '{self.alias}'"
                )

        if passphrase is None:
            raise AccountsError("Passphrase can't be 'None'")

        # Rest of the code to unlock the account using the passphrase
        self.__cached_signer = ApeSigner(private_key=self.__decrypt_keyfile(passphrase))
        self.locked = False

    def lock(self):
        self.locked = True

    def change_password(self):
        self.locked = True  # force entering passphrase to get key
        key = self.__signer.private_key

        passphrase = self._prompt_for_passphrase("Create new passphrase", confirmation_prompt=True)
        self.keyfile_path.write_text(
            json.dumps(EthAccount.encrypt(key, passphrase)), encoding="utf8"
        )

    def delete(self):
        passphrase = self._prompt_for_passphrase(
            f"Enter passphrase to delete '{self.alias}'", default=""
        )
        self.__decrypt_keyfile(passphrase)
        self.keyfile_path.unlink()

    def sign_authorization(
        self,
        address: AddressType,
        chain_id: int | None = None,
        nonce: int | None = None,
    ) -> MessageSignature | None:
        if chain_id is None:
            chain_id = self.provider.chain_id

        display_msg = f"Allow **full root access** to '{address}' on {chain_id or 'any chain'}?"
        if not click.confirm(click.style(f"{display_msg}\n\nAcknowledge: ", fg="yellow")):
            return None

        return self.__signer.sign_authorization(address=address, chain_id=chain_id, nonce=nonce)

    def sign_message(self, msg: Any, **signer_options) -> MessageSignature | None:
        display_msg, msg = _get_signing_message_with_display(msg)
        if display_msg is None:
            logger.warning("Unsupported message type, (type=%r, msg=%r)", type(msg), msg)
            return None

        if self.__autosign or click.confirm(f"{display_msg}\n\nSign: "):
            return self.__signer.sign_message(msg, **signer_options)

        return None

    def sign_transaction(self, txn: "TransactionAPI", **signer_options) -> "TransactionAPI | None":
        if not (self.__autosign or click.confirm(f"{txn}\n\nSign: ")):
            return None

        return self.__signer.sign_transaction(txn, **signer_options)

    def sign_raw_msghash(self, msghash: HexBytes) -> MessageSignature | None:
        logger.warning(
            "Signing a raw hash directly is a dangerous action which could risk "
            "substantial losses! Only confirm if you are 100% sure of the origin!"
        )

        # NOTE: Signing a raw hash is so dangerous, we don't want to allow autosigning it
        if not click.confirm("Please confirm you wish to sign using `EthAccount.signHash`"):
            return None

        # Ignoring misleading deprecated warning from web3.py.
        # Also, we have already warned the user about the safety.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return self.__signer.sign_raw_msghash(msghash)

    def set_autosign(self, enabled: bool, passphrase: str | None = None):
        """
        Allow this account to automatically sign messages and transactions.

        Args:
            enabled (bool): ``True`` to enable, ``False`` to disable.
            passphrase (str | None): Optionally provide the passphrase.
              If not provided, you will be prompted to enter it.
        """
        if enabled:
            self.unlock(passphrase=passphrase)
            logger.warning("Danger! This account will now sign any transaction it's given.")

        self.__autosign = enabled
        if not enabled:
            # Re-lock if it was turning off
            self.locked = True
            self.__cached_signer = None

    def _prompt_for_passphrase(self, message: str | None = None, **kwargs) -> str:
        message = message or f"Enter passphrase to unlock '{self.alias}'"
        return click.prompt(
            message,
            hide_input=True,
            **kwargs,
        )

    def __decrypt_keyfile(self, passphrase: str) -> bytes:
        try:
            return EthAccount.decrypt(self.keyfile, passphrase)
        except ValueError as err:
            raise InvalidPasswordError() from err

    def set_delegate(self, contract: BaseAddress | AddressType | str, **txn_kwargs):
        return self.__signer.set_delegate(contract, **txn_kwargs)

    def remove_delegate(self, **txn_kwargs):
        return self.__signer.remove_delegate(**txn_kwargs)


def _write_and_return_account(
    alias: str, passphrase: str, account: "LocalAccount"
) -> KeyfileAccount:
    """Write an account to disk and return an Ape KeyfileAccount"""
    path = ManagerAccessMixin.account_manager.containers["accounts"].data_folder.joinpath(
        f"{alias}.json"
    )
    path.write_text(json.dumps(EthAccount.encrypt(account.key, passphrase)), encoding="utf8")

    return KeyfileAccount(keyfile_path=path)


def generate_account(
    alias: str, passphrase: str, hd_path: str = ETHEREUM_DEFAULT_PATH, word_count: int = 12
) -> tuple[KeyfileAccount, str]:
    """
    Generate a new account.

    Args:
        alias (str): The alias name of the account.
        passphrase (str): Passphrase used to encrypt the account storage file.
        hd_path (str): The hierarchical deterministic path to use when generating the account.
            Defaults to `m/44'/60'/0'/0/0`.
        word_count (int): The amount of words to use in the generated mnemonic.

    Returns:
        Tuple of :class:`~ape_accounts.accounts.KeyfileAccount` and mnemonic for the generated
        account.
    """
    EthAccount.enable_unaudited_hdwallet_features()

    alias = _validate_account_alias(alias)
    passphrase = _validate_account_passphrase(passphrase)

    account, mnemonic = EthAccount.create_with_mnemonic(num_words=word_count, account_path=hd_path)
    ape_account = _write_and_return_account(alias, passphrase, account)
    return ape_account, mnemonic


def import_account_from_mnemonic(
    alias: str, passphrase: str, mnemonic: str, hd_path: str = ETHEREUM_DEFAULT_PATH
) -> KeyfileAccount:
    """
    Import a new account from a mnemonic seed phrase.

    Args:
        alias (str): The alias name of the account.
        passphrase (str): Passphrase used to encrypt the account storage file.
        mnemonic (str): List of space-separated words representing the mnemonic seed phrase.
        hd_path (str): The hierarchical deterministic path to use when generating the account.
            Defaults to `m/44'/60'/0'/0/0`.

    Returns:
        Tuple of AccountAPI and mnemonic for the generated account.
    """
    EthAccount.enable_unaudited_hdwallet_features()

    alias = _validate_account_alias(alias)
    passphrase = _validate_account_passphrase(passphrase)

    account = EthAccount.from_mnemonic(mnemonic, account_path=hd_path)

    return _write_and_return_account(alias, passphrase, account)


def import_account_from_private_key(
    alias: str, passphrase: str, private_key: str
) -> KeyfileAccount:
    """
    Import a new account from a mnemonic seed phrase.

    Args:
        alias (str): The alias name of the account.
        passphrase (str): Passphrase used to encrypt the account storage file.
        private_key (str): Hex string private key to import.

    Returns:
        Tuple of AccountAPI and mnemonic for the generated account.
    """
    alias = _validate_account_alias(alias)
    passphrase = _validate_account_passphrase(passphrase)

    account = EthAccount.from_key(to_bytes(hexstr=private_key))

    return _write_and_return_account(alias, passphrase, account)


# Abstracted to make testing easier.
def _get_signing_message_with_display(msg) -> tuple[str | None, Any]:
    display_msg = None

    if isinstance(msg, str):
        display_msg = f"Signing raw string: '{msg}'"
        msg = encode_defunct(text=msg)

    elif isinstance(msg, int):
        display_msg = f"Signing raw integer: {msg}"
        msg = encode_defunct(hexstr=to_hex(msg))

    elif isinstance(msg, bytes):
        display_msg = f"Signing raw bytes: '{to_hex(msg)}'"
        msg = encode_defunct(primitive=msg)

    elif isinstance(msg, EIP712Message):
        # Display message data to user
        display_msg = "Signing EIP712 Message\n"

        # Domain Data
        display_msg += "Domain\n"
        if msg._eip712_domain_.name:
            display_msg += f"\tName: {msg._eip712_domain_.name}\n"
        if msg._eip712_domain_.version:
            display_msg += f"\tVersion: {msg._eip712_domain_.version}\n"
        if msg._eip712_domain_.chainId:
            display_msg += f"\tChain ID: {msg._eip712_domain_.chainId}\n"
        if msg._eip712_domain_.verifyingContract:
            display_msg += f"\tContract: {msg._eip712_domain_.verifyingContract}\n"
        if msg._eip712_domain_.salt:
            display_msg += f"\tSalt: {to_hex(msg._eip712_domain_.salt)}\n"

        # Message Data
        display_msg += "Message\n"
        for field, value in msg.model_dump().items():
            if isinstance(value, dict):
                msg_value = ""
                for sub_field, attr in value.items():
                    if isinstance(attr, bytes):
                        attr = to_hex(attr)

                    msg_value += f"\t\t{sub_field}: {attr}\n"

                display_msg += f"\t{field}:\n{msg_value}\n"

            else:
                if isinstance(value, bytes):
                    value = to_hex(value)

                display_msg += f"\t{field}: {value}\n"

        display_msg = display_msg.strip()
        msg = msg.signable_message

    elif isinstance(msg, SignableMessage):
        display_msg = f"{msg}"

    # Using 2 spaces is cleaner than a full tab.
    if display_msg is not None:
        display_msg = display_msg.replace("\t", "  ")

    return (display_msg, msg)



import copy
from pathlib import Path

import pytest

import ape
from ape.api.networks import EcosystemAPI
from ape.api.providers import SubprocessProvider
from ape.exceptions import NetworkError, ProviderNotFoundError
from ape.managers.networks import NodeProcessData, NodeProcessMap
from ape.utils.misc import LOCAL_NETWORK_NAME
from ape.utils.os import create_tempdir
from ape.utils.testing import DEFAULT_TEST_CHAIN_ID


class NewChainID:
    chain_id = 12345678987654

    def __call__(self, *args, **kwargs) -> int:
        self.chain_id += 1
        return self.chain_id


chain_id_factory = NewChainID()

DEFAULT_CHOICES = {
    "::node",
    "::test",
    ":sepolia",
    ":sepolia:node",
    ":local",
    ":mainnet",
    ":mainnet:node",
    "ethereum",
    "ethereum::test",
    "ethereum::node",
    "ethereum:sepolia",
    "ethereum:sepolia:node",
    "ethereum:local",
    "ethereum:local:node",
    "ethereum:local:test",
    "ethereum:mainnet",
    "ethereum:mainnet:node",
}


@pytest.fixture
def get_provider_with_unused_chain_id(networks_connected_to_tester):
    networks = networks_connected_to_tester

    def fn(**more_settings):
        chain_id = chain_id_factory()
        settings = {"chain_id": chain_id, **more_settings}
        choice = "ethereum:local:test"
        disconnect_after = settings.pop("disconnect_after", True)
        context = networks.parse_network_choice(
            choice, disconnect_after=disconnect_after, provider_settings=settings
        )
        return context

    return fn


@pytest.fixture
def get_context(networks_connected_to_tester):
    def fn():
        return networks_connected_to_tester.parse_network_choice("ethereum:local:test")

    return fn


@pytest.fixture
def network_with_no_providers(ethereum):
    network = ethereum.get_network("sepolia-fork")
    default_provider = network.default_provider
    providers = network.__dict__["providers"]

    if default_provider or providers:
        # Handle user running tests with forked-network plugins installed
        network._default_provider = None
        network.__dict__["providers"] = []

    yield network

    if default_provider or providers:
        network._default_provider = default_provider.name
        network.__dict__["providers"] = providers
        assert network.default_provider, "Tear-down failed - providers not set."


def test_get_network_choices(networks, ethereum, mocker):
    # Simulate having a provider like foundry installed.
    mock_provider = mocker.MagicMock()
    mock_provider.name = "mock"
    ethereum.networks["mainnet-fork"].providers["mock"] = mock_provider
    ethereum.networks["local"].providers["mock"] = mock_provider

    # Ensure the provider shows up both mainnet fork and local
    # (There was once a bug where it was skipped!)
    expected = {":mainnet-fork:mock", ":local:mock"}

    actual = {c for c in networks.get_network_choices()}
    expected = DEFAULT_CHOICES.union(expected)
    assert expected.issubset(actual)


def test_get_network_choices_filter_ecosystem(networks):
    actual = {c for c in networks.get_network_choices(ecosystem_filter="ethereum")}
    assert DEFAULT_CHOICES.issubset(actual)


def test_get_network_choices_filter_network(networks):
    actual = {c for c in networks.get_network_choices(network_filter="mainnet")}
    mainnet_choices = {
        ":mainnet",
        ":mainnet:node",
        "ethereum",
        "ethereum:mainnet",
        "ethereum:mainnet:node",
    }
    assert mainnet_choices.issubset(actual)


def test_get_network_choices_filter_provider(networks):
    actual = {c for c in networks.get_network_choices(provider_filter="test")}
    expected = {"::test", ":local", "ethereum:local", "ethereum:local:test", "ethereum"}
    assert all(e in actual for e in expected)


def test_get_provider_when_no_default(network_with_no_providers):
    expected = f"No default provider for network '{network_with_no_providers.name}'"
    with pytest.raises(NetworkError, match=expected):
        # Not provider installed out-of-the-box for sepolia-fork network
        provider = network_with_no_providers.get_provider()
        assert not provider, f"Provider should be None but got '{provider.name}'"


def test_repr_connected_to_local(networks_connected_to_tester):
    actual = repr(networks_connected_to_tester)
    expected = f"<NetworkManager active_provider=<Test chain_id={DEFAULT_TEST_CHAIN_ID}>>"
    assert actual == expected

    # Check individual network
    actual = repr(networks_connected_to_tester.provider.network)
    expected = f"<ethereum:local chain_id={DEFAULT_TEST_CHAIN_ID}>"
    assert actual == expected


def test_repr_disconnected(networks_disconnected):
    assert repr(networks_disconnected) == "<NetworkManager>"
    assert repr(networks_disconnected.ethereum) == "<ethereum>"
    assert repr(networks_disconnected.ethereum.local) == "<ethereum:local>"
    assert repr(networks_disconnected.ethereum.sepolia) == "<ethereum:sepolia chain_id=11155111>"


def test_get_provider_from_choice_custom_provider(networks_connected_to_tester):
    uri = "https://geth:1234567890abcdef@geth.foo.bar/"
    provider = networks_connected_to_tester.get_provider_from_choice(f"ethereum:local:{uri}")
    assert uri in provider.connection_id
    assert provider.name == "node"
    assert provider.uri == uri
    assert provider.network.name == "local"  # Network was specified to be local!
    assert provider.network.ecosystem.name == "ethereum"


def test_get_provider_from_choice_custom_adhoc_ecosystem(networks_connected_to_tester):
    uri = "https://geth:1234567890abcdef@geth.foo.bar/"
    provider = networks_connected_to_tester.get_provider_from_choice(uri)
    assert provider.name == "node"
    assert provider.uri == uri
    assert provider.network.name == "custom"
    assert provider.network.ecosystem.name == "ethereum"


def test_parse_network_choice_same_provider(chain, networks_connected_to_tester, get_context):
    context = get_context()
    start_count = len(context.connected_providers)
    original_block_number = chain.blocks.height
    object_id = id(chain.provider)

    with context:
        assert id(chain.provider) == object_id
        count = len(context.connected_providers)

        # Does not create a new provider since it is the same chain ID
        assert count == start_count

    assert id(chain.provider) == object_id
    assert len(context.connected_providers) == start_count
    assert chain.blocks.height == original_block_number

    for provider in context.connected_providers.values():
        assert provider._web3 is not None


@pytest.mark.xdist_group(name="multiple-eth-testers")
def test_parse_network_choice_new_chain_id(get_provider_with_unused_chain_id, get_context):
    start_count = len(get_context().connected_providers)
    context = get_provider_with_unused_chain_id()
    with context:
        count = len(context.connected_providers)

        assert context._provider.chain_id != DEFAULT_TEST_CHAIN_ID

        # Creates new provider since it has a new chain ID
        assert count == start_count + 1

    for provider in context.connected_providers.values():
        assert provider._web3 is not None


@pytest.mark.xdist_group(name="multiple-eth-testers")
def test_parse_network_choice_disconnect_after(get_provider_with_unused_chain_id):
    context = get_provider_with_unused_chain_id()
    with context as provider:
        connection_id = provider.connection_id
        assert connection_id in context.connected_providers

    assert connection_id not in context.connected_providers


@pytest.mark.xdist_group(name="multiple-eth-testers")
def test_parse_network_choice_multiple_contexts(
    eth_tester_provider, get_provider_with_unused_chain_id
):
    first_context = get_provider_with_unused_chain_id()
    assert eth_tester_provider.chain_id == DEFAULT_TEST_CHAIN_ID, (
        "Test setup failed - expecting to start on default chain ID"
    )
    assert eth_tester_provider.make_request("eth_chainId") == DEFAULT_TEST_CHAIN_ID

    with first_context:
        start_count = len(first_context.connected_providers)
        expected_next_count = start_count + 1
        second_context = get_provider_with_unused_chain_id()
        with second_context:
            # Second context should already know about connected providers
            assert len(first_context.connected_providers) == expected_next_count
            assert len(second_context.connected_providers) == expected_next_count

    assert eth_tester_provider.chain_id == DEFAULT_TEST_CHAIN_ID
    assert eth_tester_provider.make_request("eth_chainId") == DEFAULT_TEST_CHAIN_ID


def test_getattr_ecosystem_with_hyphenated_name(networks, ethereum):
    networks._plugin_ecosystems["hyphen-in-name"] = networks.ecosystems["ethereum"]
    assert networks.hyphen_in_name  # Make sure does not raise AttributeError
    del networks.ecosystems["hyphen-in-name"]


def test_getattr_custom_ecosystem(networks, custom_networks_config_dict, project):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["ecosystem"] = "custom-ecosystem"

    with project.temp_config(**data):
        actual = networks.custom_ecosystem
        assert isinstance(actual, EcosystemAPI)


def test_getattr_plugin_ecosystem_same_as_evm_chains(mocker, networks):
    """
    Simulated having a plugin ecosystem installed.
    """
    optimismy = mocker.MagicMock()
    networks._plugin_ecosystems["optimismy"] = optimismy
    actual = networks.optimismy
    del networks._plugin_ecosystems["optimismy"]
    assert actual == optimismy


def test_getattr_evm_chains_ecosystem(networks):
    """
    Show we can getattr evm-chains only ecosystems.
    """
    actual = networks.moonbeam
    assert actual.name == "moonbeam"


def test_getattr_plugin_ecosystem_same_name_as_evm_chains(mocker, networks):
    """
    Show when an ecosystem is both in evm-chains and an installed plugin
    that Ape prefers the installed plugin.
    """
    moonbeam_plugin = mocker.MagicMock()
    networks._plugin_ecosystems["moonbeam"] = moonbeam_plugin
    actual = networks.moonbeam
    del networks._plugin_ecosystems["moonbeam"]
    assert actual == moonbeam_plugin
    assert actual != networks._evmchains_ecosystems["moonbeam"]


@pytest.mark.parametrize("scheme", ("http", "https"))
def test_create_custom_provider_http(networks, scheme):
    provider = networks.create_custom_provider(f"{scheme}://example.com")
    assert provider.uri == f"{scheme}://example.com"


@pytest.mark.parametrize("scheme", ("ws", "wss"))
def test_create_custom_provider_ws(networks, scheme):
    with pytest.raises(NetworkError):
        networks.create_custom_provider(f"{scheme}://example.com")


def test_create_custom_provider_ipc(networks):
    provider = networks.create_custom_provider("path/to/geth.ipc")
    assert provider.ipc_path == Path("path/to/geth.ipc")
    assert provider.uri == f"{provider.ipc_path}"


def test_ecosystems(networks):
    actual = networks.ecosystems
    assert "ethereum" in actual
    assert actual["ethereum"].name == "ethereum"


def test_ecosystems_include_custom(networks, custom_networks_config_dict, project):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["ecosystem"] = "custom-ecosystem"
    with project.temp_config(**data):
        actual = networks.ecosystems

    assert "custom-ecosystem" in actual


def test_ecosystems_when_custom_has_bad_base_ecosystem(
    networks, custom_networks_config_dict, project
):
    data = copy.deepcopy(custom_networks_config_dict)
    eco_name = "custom-ecosystem-bad"
    eco_plugin_name = "plugin-ecosystem-not-installed"
    # Only hits error if ecosystem is custom because it is unknown.
    data["networks"]["custom"][0]["ecosystem"] = eco_name
    # The plugin base is for the class-implementation. If it is not a real plugin,
    # it should fail nicely.
    data["networks"]["custom"][0]["base_ecosystem_plugin"] = eco_plugin_name
    expected = (
        rf"Custom network '{eco_name}:apenet' specified unknown "
        rf"base-ecosystem class '{eco_plugin_name}'\. "
        rf"Are you missing plugin 'ape-{eco_plugin_name}'\?"
    )
    with project.temp_config(**data):
        with pytest.raises(NetworkError, match=expected):
            _ = networks.ecosystems


def test_fork_network_not_forkable(networks, eth_tester_provider):
    """
    Show correct failure when trying to fork the local network.
    """
    expected = "Unable to fork network 'local'."
    with pytest.raises(NetworkError, match=expected):
        with networks.fork():
            pass


def test_fork_no_providers(networks, mock_sepolia, disable_fork_providers):
    """
    Show correct failure when trying to fork without
    ape-hardhat or ape-foundry installed.
    """
    expected = "No providers for network 'sepolia-fork'."
    with pytest.raises(NetworkError, match=expected):
        with networks.fork():
            pass


def test_fork_use_non_existing_provider(networks, mock_sepolia):
    """
    Show correct failure when specifying a non-existing provider.
    """
    expected = "No provider named 'NOT_EXISTS' in network 'sepolia-fork' in ecosystem 'ethereum'.*"
    with pytest.raises(ProviderNotFoundError, match=expected):
        with networks.fork(provider_name="NOT_EXISTS"):
            pass


def test_fork(networks, mock_sepolia, mock_fork_provider):
    """
    Happy-path fork test.
    """
    ctx = networks.fork()
    assert ctx._disconnect_after is True
    with ctx as provider:
        assert provider.name == "mock"
        assert provider.network.name == "sepolia-fork"
        # Fork the fork.
        ctx2 = networks.fork()
        with ctx2 as provider2:
            assert provider2.partial_call[1]["provider_settings"]["host"] == "auto"
            assert provider2.name == "mock"
            assert provider2.network.name == "sepolia-fork"


def test_fork_specify_provider(networks, mock_sepolia, mock_fork_provider):
    """
    Happy-path fork test when specifying the provider.
    """
    ctx = networks.fork(provider_name="mock")
    assert ctx._disconnect_after is True
    with ctx as provider:
        assert provider.name == "mock"
        assert provider.network.name == "sepolia-fork"


def test_fork_with_provider_settings(networks, mock_sepolia, mock_fork_provider):
    """
    Show it uses the given provider settings.
    """
    settings = {"fork": {"ethereum": {"sepolia": {"block_number": 123}}}}
    with networks.fork(provider_settings=settings):
        actual = mock_fork_provider.partial_call
        assert actual[1]["provider_settings"] == settings


def test_fork_with_positive_block_number(networks, mock_sepolia, mock_fork_provider):
    block_id = 123
    with networks.fork(block_number=block_id):
        call = mock_fork_provider.partial_call

    settings = call[1]["provider_settings"]["fork"]["ethereum"]["sepolia"]
    actual = settings["block_number"]
    assert actual == block_id


def test_fork_with_negative_block_number(
    networks, mock_sepolia, mock_fork_provider, eth_tester_provider
):
    # Mine so we are past genesis.
    block_id = -1
    block = eth_tester_provider.get_block("latest")
    mock_fork_provider.get_block.return_value = block

    with networks.fork(block_number=block_id):
        call = mock_fork_provider.partial_call

    actual = call[1]["provider_settings"]["fork"]["ethereum"]["sepolia"]["block_number"]
    expected = block.number - 1  # Relative to genesis!
    assert actual == expected


def test_fork_past_genesis(networks, mock_sepolia, mock_fork_provider, eth_tester_provider):
    block_id = -10_000_000_000
    with pytest.raises(NetworkError, match="Unable to fork past genesis block."):
        with networks.fork(block_number=block_id):
            pass


def test_getitem(networks):
    ethereum = networks["ethereum"]
    assert ethereum.name == "ethereum"
    assert isinstance(ethereum, EcosystemAPI)


def test_network_names(networks, custom_networks_config_dict, project):
    data = copy.deepcopy(custom_networks_config_dict)
    with project.temp_config(**data):
        actual = networks.network_names
        assert LOCAL_NETWORK_NAME in actual
        assert "mainnet" in actual  # Normal
        assert "mainnet-fork" in actual  # Forked
        assert "apenet" in actual  # Custom
        assert "apenet-fork" in actual  # Custom forked


def test_custom_networks_defined_in_non_local_project(custom_networks_config_dict):
    """
    Showing we can read and use custom networks that are not defined
    in the local project.
    """
    # Ensure we are using a name that is not used anywhere else, for accurte testing.
    net_name = "customnetdefinedinnonlocalproj"
    eco_name = "customecosystemnotdefinedyet"
    custom_networks = copy.deepcopy(custom_networks_config_dict)
    custom_networks["networks"]["custom"][0]["name"] = net_name
    custom_networks["networks"]["custom"][0]["ecosystem"] = eco_name

    with ape.Project.create_temporary_project(config_override=custom_networks) as temp_project:
        nm = temp_project.network_manager

        # Tests `.get_ecosystem()` for custom networks.
        ecosystem = nm.get_ecosystem(eco_name)
        assert ecosystem.name == eco_name

        network = ecosystem.get_network(net_name)
        assert network.name == net_name


def test_get_ecosystem(networks):
    ethereum = networks.get_ecosystem("ethereum")
    assert isinstance(ethereum, EcosystemAPI)
    assert ethereum.name == "ethereum"


def test_get_ecosystem_from_evmchains(networks):
    """
    Show we can call `.get_ecosystem()` for an ecosystem only
    defined in evmchains.
    """
    moonbeam = networks.get_ecosystem("moonbeam")
    assert isinstance(moonbeam, EcosystemAPI)
    assert moonbeam.name == "moonbeam"


class TestNodeProcessData:
    def test_matches_provider(self, eth_tester_provider):
        data = NodeProcessData(
            network_choice=f"{eth_tester_provider.network.choice}:{eth_tester_provider.name}",
            ipc_path="test.ipc",
        )
        assert not data.matches_provider(eth_tester_provider)
        data.ipc_path = None
        assert data.matches_provider(eth_tester_provider)


class TestNodeProcessMap:
    def test_cache_and_remove_provider(self, mocker, eth_tester_provider):
        mock_process = mocker.MagicMock()
        mock_process.pid = 12345678901234567890

        class MyFakeProvider(SubprocessProvider):
            @property
            def is_connected(self):
                return True

        # Hack to allow abstract methods.
        MyFakeProvider.__abstractmethods__ = set()  # type: ignore
        provider = MyFakeProvider(name="fake", network=eth_tester_provider.network)  # type: ignore
        provider.process = mock_process

        with create_tempdir() as tmp:
            file = tmp / "networks.json"
            mapping = NodeProcessMap(path=file)
            mapping.cache_provider(provider)
            assert provider in mapping

            # Not test removing it.
            mapping.remove_provider(provider)
            assert provider not in mapping



def __getattr__(name: str):
    if name in (
        "AccountAPI",
        "AccountContainerAPI",
        "ImpersonatedAccount",
        "TestAccountAPI",
        "TestAccountContainerAPI",
    ):
        import ape.api.accounts as accounts_module

        return getattr(accounts_module, name)

    elif name in ("Address",):
        import ape.api.address as address_module

        return getattr(address_module, name)

    elif name in ("CompilerAPI",):
        import ape.api.compiler as compiler_module

        return getattr(compiler_module, name)

    elif name in ("ConfigDict", "ConfigEnum", "PluginConfig"):
        import ape.api.config as config_module

        return getattr(config_module, name)

    elif name in ("ConverterAPI",):
        import ape.api.convert as convert_module

        return getattr(convert_module, name)

    elif name in ("ExplorerAPI",):
        import ape.api.explorers as explorer_module

        return getattr(explorer_module, name)

    elif name in ("BlockAPI, ProviderAPI, SubprocessProvider, TestProviderAPI, UpstreamProvider"):
        import ape.api.providers as provider_module

        return getattr(provider_module, name)

    elif name in (
        "EcosystemAPI",
        "ForkedNetworkAPI",
        "NetworkAPI",
        "ProviderContextManager",
        "create_network_type",
    ):
        import ape.api.networks as network_module

        return getattr(network_module, name)

    elif name in ("DependencyAPI", "ProjectAPI"):
        import ape.api.projects as project_module

        return getattr(project_module, name)

    elif name in ("QueryAPI", "QueryType"):
        import ape.api.query as query_module

        return getattr(query_module, name)

    elif name in ("TraceAPI",):
        import ape.api.trace as trace_module

        return getattr(trace_module, name)

    elif name in ("ReceiptAPI", "TransactionAPI"):
        import ape.api.transactions as tx_module

        return getattr(tx_module, name)

    else:
        raise AttributeError(name)


__all__ = [
    "AccountAPI",
    "AccountContainerAPI",
    "Address",
    "BlockAPI",
    "CompilerAPI",
    "ConfigDict",
    "ConfigEnum",
    "ConverterAPI",
    "DependencyAPI",
    "EcosystemAPI",
    "ExplorerAPI",
    "ForkedNetworkAPI",
    "ImpersonatedAccount",
    "NetworkAPI",
    "PluginConfig",
    "ProjectAPI",
    "ProviderAPI",
    "ProviderContextManager",
    "QueryAPI",
    "QueryType",
    "ReceiptAPI",
    "SubprocessProvider",
    "TestAccountAPI",
    "TestAccountContainerAPI",
    "TestProviderAPI",
    "TraceAPI",
    "TransactionAPI",
    "UpstreamProvider",
    "create_network_type",
]



import shutil
from pathlib import Path

import click
from click import BadParameter

from ape.cli.options import ape_cli_context

GITIGNORE_CONTENT = """
# Ape stuff
.build/
.cache/

# Python
.env
.venv
.pytest_cache
.python-version
__pycache__
"""


def validate_github_repo(ctx, param, value):
    if not value:
        return None

    elif value.count("/") != 1:
        raise BadParameter("Invalid GitHub parameter, must be in format 'org/repo'.", ctx, param)

    org, repo = value.split("/")
    if not org:
        raise BadParameter("Invalid GitHub parameter, missing 'org' value.")
    if not repo:
        raise BadParameter("Invalid GitHub parameter, missing 'repo' value.")

    return (org, repo)


@click.command(short_help="Initialize an ape project")
@ape_cli_context()
@click.option(
    "--github",
    metavar="github-org/repo",
    help="Clone a template from Github",
    callback=validate_github_repo,
)
@click.option("--name", "project_name", prompt=True, help="A project name")
def cli(cli_ctx, github, project_name):
    """
    ``ape init`` allows the user to create an ape project with
    default folders and ape-config.yaml.
    """
    if github:
        from ape.utils._github import github_client

        org, repo = github
        github_client.clone_repo(org, repo, Path.cwd())
        shutil.rmtree(Path.cwd() / ".git", ignore_errors=True)
    else:
        project_folder = Path.cwd()

        for folder_name in ("contracts", "tests", "scripts"):
            # Create target Directory
            folder = project_folder / folder_name
            if folder.exists():
                cli_ctx.logger.warning(f"'{folder}' exists")
            else:
                folder.mkdir()

        git_ignore_path = project_folder / ".gitignore"
        if git_ignore_path.exists():
            cli_ctx.logger.warning(f"Unable to create .gitignore: '{git_ignore_path}' file exists.")
        else:
            git_ignore_path.touch()
            git_ignore_path.write_text(GITIGNORE_CONTENT.lstrip(), encoding="utf8")

        ape_config = project_folder / "ape-config.yaml"
        if ape_config.exists():
            cli_ctx.logger.warning(f"'{ape_config}' exists")
        else:
            ape_config.write_text(f"name: {project_name}\n", encoding="utf8")
            cli_ctx.logger.success(f"{project_name} is written in ape-config.yaml")



from collections.abc import Callable
from pathlib import Path

import pytest

__projects_directory__ = Path(__file__).parent / "projects"
__project_names__ = [
    p.stem for p in __projects_directory__.iterdir() if p.is_dir() and not p.stem.startswith(".")
]


def assert_failure(result, expected_output):
    assert result.exit_code == 1
    assert result.exception is not None
    assert "ERROR" in result.output
    assert expected_output in result.output


class NodeId:
    """
    A class that extracts a callable test's name and module name.
    """

    def __init__(self, test_method: Callable):
        self.module_full_name = test_method.__module__
        self.name = test_method.__name__

    @property
    def module_name(self) -> str:
        return self.module_full_name.split(".")[-1]

    @property
    def node_id(self) -> str:
        return f"{self.module_name}.{self.name}"


class ProjectSkipper:
    """
    A class that contains stateful information about
    which projects to skip tests for.
    """

    def __init__(self):
        self.projects: dict[str, dict] = {n: {} for n in __project_names__}

    def __iter__(self):
        return iter(self.projects)

    def do_skip(self, project: str, module: str, test: str) -> bool:
        """
        Returns ``True`` if a test has been marked to be
        skipped for the given project using the ``skip_project`` or
        ``skip_project_except`` decorators.
        """
        if project not in self.projects:
            # Not a project-based integration test
            return False

        result = test in self.projects[project].get(module, [])
        return result

    def _raise_if_not_exists(self, project_name: str, node_id: str):
        if project_name not in self.projects:
            raise pytest.Collector.CollectError(
                f"Project '{project_name}' does not exist (test={node_id}."
            )

    def skip_projects(self, method: Callable, *projects: str):
        """
        Call this method to record a 'skip'.
        The ``skip_project`` decorator calls this method
        on the test method they are wrapped around.
        """
        assert hasattr(method, "__name__") and hasattr(method, "__module__")
        node = NodeId(method)
        for project in projects:
            self._raise_if_not_exists(project, node.node_id)
            if node.module_name not in self.projects[project]:
                self.projects[project][node.module_name] = set()

            self.projects[project][node.module_name].add(node.name)

    def skip_projects_except(self, method: Callable, *projects: str):
        """
        Call this method to record 'skip's for each project that is not
        in the given list. The ``skip_project_except`` decorator calls
        this method on the test method they are wrapped around.
        """
        assert hasattr(method, "__name__") and hasattr(method, "__module__")
        node = NodeId(method)

        # Verify projects to run for exist
        for proj in projects:
            self._raise_if_not_exists(proj, node.node_id)

        filtered_projects = [p for p in self.projects if p not in projects]
        self.skip_projects(method, *filtered_projects)


project_skipper = ProjectSkipper()


def skip_projects(*names: str):
    """
    Use this decorator to cause a CLI integration test
    not to run for the given projects.
    """

    def decorator(f):
        project_skipper.skip_projects(f, *names)
        return f

    return decorator


def skip_projects_except(*names: str):
    """
    Use this decorator to cause a CLI integration test
    to only run for the given projects.
    """

    def decorator(f):
        project_skipper.skip_projects_except(f, *names)
        return f

    return decorator


run_once = skip_projects_except("test")
"""
For times when the CLI integration test is unlikely to be
affected by project structure.
"""


def github_xfail():
    """
    Expects failures because of GitHib rate limiting
    """

    def wrapper(f):
        f = pytest.mark.xfail(
            strict=False, reason="GitHub rate limiting issues or plugin install issues"
        )(f)
        f = run_once(f)
        return f

    return wrapper



from eth_pydantic_types import HexBytes

from ape_ethereum.proxies import ProxyType
from tests.conftest import geth_process_test


@geth_process_test
def test_standard_proxy(project, owner, geth_contract, ethereum):
    target = geth_contract.address
    contract_instance = owner.deploy(project.ERC1967Proxy, target, HexBytes(""))
    actual = ethereum.get_proxy_info(contract_instance.address)
    assert actual is not None
    assert actual.type == ProxyType.Standard
    assert actual.target == target


@geth_process_test
def test_beacon_proxy(project, geth_contract, owner, ethereum):
    target = geth_contract.address
    beacon_instance = owner.deploy(project.beacon, target)
    beacon = beacon_instance.address

    contract_instance = owner.deploy(project.BeaconProxy, beacon, HexBytes(""))

    actual = ethereum.get_proxy_info(contract_instance.address)
    assert actual is not None
    assert actual.type == ProxyType.Beacon
    assert actual.target == target


@geth_process_test
def test_uups_proxy(project, geth_contract, owner, ethereum):
    target = geth_contract.address
    contract_instance = owner.deploy(project.Uups, HexBytes("0x2beb1711"), target)
    actual = ethereum.get_proxy_info(contract_instance.address)
    assert actual is not None
    assert actual.type == ProxyType.UUPS
    assert actual.target == target


@geth_process_test
def test_gnosis_safe(project, geth_contract, owner, ethereum, chain):
    # Setup a proxy contract.
    target = geth_contract.address
    proxy_instance = owner.deploy(project.SafeProxy, target)

    # (test)
    actual = ethereum.get_proxy_info(proxy_instance.address)
    assert actual is not None
    assert actual.type == ProxyType.GnosisSafe
    assert actual.target == target

    # Ensure we can call the proxy-method.
    assert proxy_instance.masterCopy()

    # Ensure we can call target methods.
    assert isinstance(proxy_instance.myNumber(), int)

    # Ensure this works with new instances.
    proxy_instance_ref_2 = chain.contracts.instance_at(proxy_instance.address)
    assert proxy_instance_ref_2.masterCopy()
    assert isinstance(proxy_instance_ref_2.myNumber(), int)


@geth_process_test
def test_openzeppelin(project, geth_contract, owner, ethereum, sender):
    constructor_contract = owner.deploy(project.SolFallbackAndReceive)
    target = geth_contract.address
    contract_instance = owner.deploy(
        project.UpgradeabilityProxy, constructor_contract.address, target
    )
    actual = ethereum.get_proxy_info(contract_instance.address)
    assert actual is not None
    assert actual.type == ProxyType.OpenZeppelin
    assert actual.target == target


@geth_process_test
def test_delegate(project, geth_contract, owner, ethereum):
    target = geth_contract.address
    contract_instance = owner.deploy(project.ERCProxy, target)
    actual = ethereum.get_proxy_info(contract_instance.address)
    assert actual is not None
    assert actual.type == ProxyType.Delegate
    assert actual.target == target
